//=============================================================================
// AdvancedMutator.
// 
// Date: 2016
// ©copyright Francesco Biscazzo. All rights reserved.
//=============================================================================
class AdvancedMutator extends Mutator;

//Forked AdvancedMutator chains.
var AdvancedMutator nextAM;
var AdvancedMutator nextDamageAM;
var AdvancedMutator nextMessageAM;
var AdvancedMutator nextHUDAM; //Remember this is a forked chain, when declaring PostRender().

var AdvancedMutatorManager AMM;

var() bool bAutoRegister; //Register this mutator as next mutator of the GameInfo.BaseMutator.
var(Optimization) int maxNoneTimes; //How many times the ActorInfo can be without a Owner before of being destroyed.
var bool bDebug; //Used to view the ActorInfo's through the use of Mutate() function.

function PostBeginPlay() {
	local bool bIsAMMInGame;
	local Mutator nextMut, lastMut;
	
	super.PostBeginPlay();
	
	if (bAutoRegister)
		Level.Game.BaseMutator.AddMutator(Self);
		
	if (maxNoneTimes <= 0)
		maxNoneTimes = 3;
	
	//Check if there's already an AdvancedMutatorManager in the map, if not, spawn it and register it to this mutator.
	foreach AllActors(class'AdvancedMutatorManager', AMM) {
		bIsAMMInGame = true;
		
		break;
	}
	
	if (!bIsAMMInGame) {
		AMM = Spawn(class'AdvancedMutatorManager');
		if (AMM != None)
			AMM.RegisterAdvancedMutator(Self);
	}
}

function ActorInfo getActorAI(Actor actor) {
	local ActorInfo AI;
	
	if (actor != None)
		foreach AllActors(class'ActorInfo', AI)
			if (AI.Owner == actor)
				return AI;
	
	return None;
}

function bool canHaveAI(Actor actor) {
	
	if (!actor.IsA('ActorInfo') && !actor.IsA('ActorInfoPickup') && !actor.IsA('ValueRegister')) {
		if (nextAM != None)
			return nextAM.canHaveAI(actor);
		else
			return true;
	}
	
	return false;
}

function bool startWithAI(Actor actor) {
	if (nextAM != None)
		return nextAM.startWithAI(actor);
	
	return false;
}

function bool canBlockActors(Actor actor) {
	if (!actor.IsA('Brush') && !actor.IsA('Projectile')) {
		if (nextAM != None)
			return nextAM.canBlockActors(actor);
	}
	
	return false;
}

function bool canBlockPlayers(Actor actor) {
	if (!actor.IsA('Brush') && !actor.IsA('Projectile')) {
		if (nextAM != None)
			return nextAM.canBlockPlayers(actor);
	}
	
	return false;
}

function bool registerTextValue(ActorInfo AI, String valueName, String value, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	return AI.registerTextValue(class, valueName, value, bCheckStatus, bUnchangedLoop, channel);
}

function bool registerOwnerValue(ActorInfo AI, String propertyName, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	return AI.registerOwnerValue(class, propertyName, bCheckStatus, bUnchangedLoop, channel);
}

function String getRegisteredValue(ActorInfo AI, String propertyName, optional int channel, optional out ValueRegister vRegister) {
	return AI.getRegisteredValue(class, propertyName, channel, vRegister);
}

function bool removeRegisteredValue(ActorInfo AI, String propertyName, optional int channel) {
	return AI.removeRegisteredValue(class, propertyName, channel);
}

//Register a text value for any actor of the specified class.
function registerClassTextValue(class<Actor> actorClass, bool bNoSubClass, String property, String value, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	local Actor a;
	
	foreach AllActors(actorClass, a)
		if (bool(ClassIsRelationedWith(a.class, actorClass)))
			registerTextValue(getActorAI(a), property, value, bCheckStatus, bUnchangedLoop, channel);
}

//Register the owner value of any actor of the specified class.
function registerClassOwnerValue(class<Actor> actorClass, bool bNoSubClass, String property, optional bool bCheckStatus, optional bool bUnchangedLoop, optional int channel) {
	local Actor a;
	
	foreach AllActors(actorClass, a)
		if (bool(ClassIsRelationedWith(a.class, actorClass)))
			registerOwnerValue(getActorAI(a), property, bCheckStatus, bUnchangedLoop, channel);
}

////////////////////////////////////////////////////////////////////////////////
//AdvancedMutator callbacks.

//Called when an AI registers to this mutator.
function AIRegistered(ActorInfo AI, Actor actor) {
	if (nextAM != None)
		nextAM.AIRegistered(AI, actor);
}

function AIDestroying(ActorInfo AI, Actor actor) {
	if (nextAM != None)
		nextAM.AIDestroying(AI, actor);
}

//Called every tick by the AI.
function TickCallback(ActorInfo AI, Actor actor) {
	if (nextAM != None)
		nextAM.TickCallback(AI, actor);
}

//Called when the owner of an AI has been None for more than "maxNoneTimes" times.
function BecameNone(ActorInfo AI) {
	if (nextAM != None)
		nextAM.BecameNone(AI);
}

//Called when a registered value changes.
function ValueStatusChanged(Actor actor, String propertyName, String oldValue, String newValue, int channel) {}

//Called when a registered value has been unchanged for one tick.
//NOTE: Called once only after the first unchanging if bUnchangedLoop is set to false.
function ValueStatusUnchanged(Actor actor, String propertyName, String value, float lastChangeTime, int channel) {}

//Called when the actors picks up an Inventory.
//NOTE: Some inventories may want to not to be pickedup if the Actor already posses a PickUp of the same class.
//An item can be some types of ammo which instead of being picked up, after the first picking up, they just increase the ammo number.
function PickedUp(Actor actor, Inventory item) {
	if (nextAM != None)
		nextAM.PickedUp(actor, item);
}

//Called on Fire and AltFire before a possibly projectile is spawned.
//Check the state of the weapon to get the fire type.
function PreFire(Actor actor, Weapon weapon, Actor HitActor) {
	if (nextAM != None)
		nextAM.PreFire(actor, weapon, HitActor);
}

//Called when a projectile is spawned.
function PostFire(Actor actor, Weapon weapon, Actor HitActor, Projectile projectile) {
	if (nextAM != None)
		nextAM.PostFire(actor, weapon, HitActor, projectile);
}

function Jumped(Actor actor) {
	if (nextAM != None)
		nextAM.Jumped(actor);
}

function ChangedWeapon(Actor actor, Weapon oldWeapon, Weapon newWeapon) {
	if (nextAM != None)
		nextAM.ChangedWeapon(actor, oldWeapon, newWeapon);
}

function RecommendingWeapon(Actor actor, out Weapon recommendedWeapon, out float recWeaponRating, out int recBUseAltMode) {
	if (nextAM != None)
		nextAM.RecommendingWeapon(actor, recommendedWeapon, recWeaponRating, recBUseAltMode);
}

function SelectingNext(Actor actor, out Inventory nextInv) {
	if (nextAM != None)
		nextAM.SelectingNext(actor, nextInv);
}

function DamageReduced(Actor actor, int Damage, out int ReducedAmount, name DamageType, vector HitLocation) {
	if (nextDamageAM != None)
		nextDamageAM.DamageReduced(actor, Damage, ReducedAmount, DamageType, HitLocation);
}

function PrioritizingArmor(Actor actor, out Inventory armor, int Damage, name DamageType, vector HitLocation) {
	if (nextDamageAM != None)
		nextDamageAM.PrioritizingArmor(actor, armor, Damage, DamageType, HitLocation);
}

function SettingDisplay(Actor actor) {
	if (nextHUDAM != None)
		nextHUDAM.SettingDisplay(actor);
}

function HeadInWater(Actor actor) {
	if (nextAM != None)
		nextAM.HeadInWater(actor);
}

function HeadOutOfWater(Actor actor) {
	if (nextAM != None)
		nextAM.HeadOutOfWater(actor);
}

function Triggered(Actor actor, Actor Other, Pawn EventInstigator) {
	if (nextAM != None)
		nextAM.Triggered(actor, Other, EventInstigator);
}

function UnTriggered(Actor actor, Actor Other, Pawn EventInstigator) {
	if (nextAM != None)
		nextAM.UnTriggered(actor, Other, EventInstigator);
}

function ZoneChanged(Actor actor, ZoneInfo oldZone, ZoneInfo newZone) {
	if (nextAM != None)
		nextAM.ZoneChanged(actor, oldZone, newZone);
}

function StateChanged(Actor actor, String oldState, String newState) {
	if (nextAM != None)
		nextAM.StateChanged(actor, oldState, newState);
}

//Called only if pawn is aiming with a weapon.
function ActorAiming(Pawn pawn, Weapon weapon, Actor oldAimedActor, Actor newAimedActor, vector oldHitLocation, vector oldHitNormal, vector oldStartTrace, vector oldEndTrace, vector newHitLocation, vector newHitNormal, vector newStartTrace, vector newEndTrace) {
	if (nextAM != None)
		nextAM.ActorAiming(pawn, weapon, oldAimedActor, newAimedActor, oldHitLocation, oldHitNormal, oldStartTrace, oldEndTrace, newHitLocation, newHitNormal, newStartTrace, newEndTrace);
}

function HittingWall( Actor actor, vector HitNormal, actor HitWall ) {
	if (nextAM != None)
		nextAM.HittingWall(actor, HitNormal, HitWall);
}

function Landing( Actor actor, vector HitNormal ) {
	if (nextAM != None)
		nextAM.Landing(actor, HitNormal);
}

//Called on Touch().
function Touching( Actor actor, Actor Other ) {
	if (nextAM != None)
		nextAM.Touching(actor, Other);
}

//Called on PostTouch().
function Touched( Actor actor, Actor Other ) {
	if (nextAM != None)
		nextAM.Touched(actor, Other);
}

function UnTouched( Actor actor, Actor Other ) {
	if (nextAM != None)
		nextAM.UnTouched(actor, Other);
}

// Called when Other bumped into actor and was blocked.
// Seems like it's actually called 4 times at once.
function Bumped( Actor actor, Actor Other ) {
	if (nextAM != None)
		nextAM.Bumped(actor, Other);
}

function ChangedBase(Actor actor) {
	if (nextAM != None)
		nextAM.ChangedBase(actor);
}

function EncroachingBy( Actor actor, actor Other ) {
	if (nextAM != None)
		nextAM.EncroachingBy(actor, Other);
}

function EncroachedOn( Actor actor, actor Other ) {
	if (nextAM != None)
		nextAM.EncroachedOn(actor, Other);
}

function FallingOutOfWorld(Actor actor) {
	if (nextAM != None)
		nextAM.FallingOutOfWorld(actor);
}

////////////////////////////////////////////////////////////////////////////////
//Reflective functions.

//TypeInfo: Type'PackageOrMap.ClassOrName'
//ObjectInfo: PackageOrMap.ClassOrName

//Get the type from a string of this format:
//Type'PackageNameOrMapName.ClassNameOrActorName'
static final function String getTypeFromString(String text) {
	if (InStr(text, "'") != -1)
		text = Left(text, InStr(text, "'") - 1);
	
	return text;
}

static final function String getPackageOrMapFromString(String text) {
	if (InStr(text, "'") != -1)
		//Has brackets, so it's a type info.
		text = Mid(text, InStr(text, "'") + 1, Len(text) - (InStr(text, "'") + 1) - 1);
	
	if (InStr(text, ".") != -1)
		text = Left(text, InStr(text, ".") - 1);
	
	return text;
}

static final function String getClassOrNameText(String text) {
	if (InStr(text, "'") != -1)
		//Has brackets, so it's a type info.
		text = Mid(text, InStr(text, "'") + 1, Len(text) - (InStr(text, "'") + 1) - 1);
	
	if (InStr(text, ".") != -1)
		text = Mid(text, InStr(text, ".") + 1);
	
	return text;
}

function Actor getActorFromText(String text, optional bool bNoCaseSensitive) {
	local String actorName;
	local Actor a;
	
	if (text != "") {
		foreach AllActors(class'Actor', a) {
			if (bNoCaseSensitive) {
				if (String(a.name) == getClassOrNameText(text))
					return a;
			} else {
				if (String(a.name) ~= getClassOrNameText(text))
					return a;
			}
		}
	}
	
	return None;
}

static final function int getIntFromString(String text) {
	local bool bNeg;
	local String posNum;
	local int num;
	
	if (InStr(text, "-") != -1) {
		bNeg = true;
		text = FindSplittedStr(text, "-", 1);
	}
	
	num = int(text);
	
	if (bNeg)
		num *= -1;
		
	return num;
}

static final function float getFloatFromString(String text) {
	local bool bNeg;
	local int intNum;
	local float floatNum, num;
	
	if (InStr(text, "-") != -1) {
		bNeg = true;
		text = FindSplittedStr(text, "-", 1);
	}
		
	intNum = int(FindSplittedStr(text, ".", 0));
	//float precision is of 6 numbers after the float point.
	floatNum = float(FindSplittedStr(text, ".", 1)) / 1000000;
	
	num = intNum + floatNum;
	
	if (bNeg)
		num *= -1;
		
	return num;
}

static final function vector getVectorFromString(String text) {
	local String XText, YText, ZText;
	local int intX, intY, intZ;
	local float floatX, floatY, floatZ;
	local int X, Y, Z;
	local bool bNegX, bNegY, bNegZ;
	local vector v;
	
	XText = FindSplittedStr(text, "X=", 1);
	XText = FindSplittedStr(XText, ",", 0);
	YText = FindSplittedStr(text, "Y=", 1);
	YText = FindSplittedStr(YText, ",", 0);
	ZText = FindSplittedStr(text, "Z=", 1);
	ZText = FindSplittedStr(ZText, ")", 0);
	
	v.X = getFloatFromString(XText);
	v.Y = getFloatFromString(YText);
	v.Z = getFloatFromString(ZText);
	
	return v;
}

static final function rotator getRotatorFromString(String text) {
	local String PitchText, YawText, RollText;
	local int Pitch, Yaw, Roll;
	local bool bNegPitch, bNegYaw, bNegRoll;
	local rotator rot;
	
	PitchText = FindSplittedStr(text, "Pitch=", 1);
	PitchText = FindSplittedStr(PitchText, ",", 0);
	YawText = FindSplittedStr(text, "Yaw=", 1);
	YawText = FindSplittedStr(YawText, ",", 0);
	RollText = FindSplittedStr(text, "Roll=", 1);
	RollText = FindSplittedStr(RollText, ")", 0);
	
	if (InStr(PitchText, "-") != -1) {
		bNegPitch = true;
		PitchText = FindSplittedStr(PitchText, "-", 1);
	}
	if (InStr(YawText, "-") != -1) {
		bNegPitch = true;
		YawText = FindSplittedStr(YawText, "-", 1);
	}
	if (InStr(RollText, "-") != -1) {
		bNegPitch = true;
		RollText = FindSplittedStr(RollText, "-", 1);
	}
	
	rot.Pitch = int(PitchText);
	rot.Yaw = int(YawText);
	rot.Roll = int(RollText);
	
	if (bNegPitch)
		rot.Pitch *= -1;
	if (bNegYaw)
		rot.Yaw *= -1;
	if (bNegRoll)
		rot.Roll *= -1;
	
	return rot;
}

//Check if an actor has the specified property, if so, if that property is of type "Actor", then return it.
//p.s. getPropertyText() returns a string of this format: ClassName'MapName.ObjectName'
function Actor getActorFromProperty(Actor actor, String propStr) {
	local String propText;

	if ((actor != None) && (propStr != "")) {
		propText = actor.getPropertyText(propStr);
		
		if (propText != "None")
			return getActorFromText(propText);
	}
	
	return None;
}

////////////////////////////////////////////////////////////////////////////////
//Gameplay functions.

//Basing on the weapon offset and to the aim of the pawn, return the aimed actor.
static final function actor getAimedActor(Pawn pawn, Weapon weapon, optional out vector HitLocation, optional out vector HitNormal, optional out vector EndTrace, optional out vector StartTrace) {
	local vector X,Y,Z;
	local float Accuracy;
	
	local rotator AdjustedAim; //To not to affect the pawn.Weapon.AdjustedAim
	
	if ((pawn != None) && (weapon != None)) {
		Accuracy = 0.0;

		GetAxes(pawn.ViewRotation, X, Y, Z);
		StartTrace = pawn.Location + weapon.CalcDrawOffset() + weapon.FireOffset.X * X + weapon.FireOffset.Y * Y + weapon.FireOffset.Z * Z; 
		AdjustedAim = pawn.AdjustAim(1000000, StartTrace, 2*weapon.AimError, False, False);	
		EndTrace = StartTrace + Accuracy * (weapon.FRand() - 0.5 )* Y * 1000
			+ Accuracy * (weapon.FRand() - 0.5 ) * Z * 1000;
		X = vector(AdjustedAim);
		EndTrace += (10000 * X);
		
		return pawn.TraceShot(HitLocation, HitNormal, EndTrace, StartTrace);
	}
	
	return None;
}

////////////////////////////////////////////////////////////////////////////////
//Miscellaneous functions.

static final function bool exists(Actor actor) {
	return actor != None;
}

//Return:
//0 - if firstClass is not relationed to secondClass.
//1 - if firstClass is the same class of secondClass.
//2 - if firstClass is child of secondClass;
//3 - if firstClass is parent of secondClass;
static final function int ClassIsRelationedWith(class firstClass, class secondClass, optional bool bNoSubClass) {
	if (firstClass == secondClass)
		return 1;
	
	if (!bNoSubClass) {
		if (ClassIsChildOf(firstClass, secondClass))
			return 2;
		else if (ClassIsChildOf(secondClass, firstClass))
			return 3;
	}
	
	return 0;
}

/* A modified version of the function "FindParameter" from CommandSystem */
/*
 * Split a string and return the string from the given position. If no string was found in the
 * given position, return an empty string.
 *
 * @str = the string to split.
 * @div = the string to interpret as divider.
 * @num = the position of the string to return.
 * @bDiv = true to keep dividers, false to remove dividers.
 */
static final function String FindSplittedStr(String str, String div, int num, optional bool bDiv) {
   local bool bEOL;
   local String tempStr;
   local int precount, curcount, wordcount, strLen;
   
   local String word;
   
   if (str != "") {
		if (InStr(str, div) != -1) {
			strLen = Len(str);
			bEOL = false;
			precount = 0;
			curcount = 0;
			wordcount = 0;
			
			while (!bEOL) {
				tempStr = Mid(str, curcount, Len(div)); //go up by div length count.
				if (tempStr != div)
					curcount++;
				else if (tempStr == div) {
					if (wordcount == num) {
						word = Mid(str, precount, curcount - precount);
						break;
					}
					wordcount++;
					if (bDiv)
						precount = curcount; //leaves the divider.
					else
						precount = curcount + Len(div); //removes the divider.
					curcount++;
				}
				
				//end of str string.
				if (curcount == strLen) {
					if (wordcount == num)
						word = Mid(str, precount);
					bEOL = true;
				}
			}
		}
		else
			word = str;
	}
	
	return word;
}

////////////////////////////////////////////////////////////////////////////////
//Forked chain functions.
//Call the standard callbacks on both the chains and the AdvancedMutator's callbacks on the AdvancedMutators only.

//Makes the mutator chain into a forked chain, so that we can call AM functions into chained AMs.
//Don't add mutators whom class is the same of an already chained mutator.
function AddMutator(Mutator M) {
	if (!M.IsA('AdvancedMutator'))
		super.AddMutator(M);
	else
		if (M.class != class)
			if (nextAM == None) {
				nextAM = AdvancedMutator(M);
			} else {
				if (nextAM.class != M.class)
					nextAM.AddMutator(M);
			}
}

function ModifyPlayer(Pawn Other) {
	super.ModifyPlayer(Other);
	if (Other != None)
		if (nextAM != None)
			nextAM.ModifyPlayer(Other);
}

function bool HandleRestartGame() {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.HandleRestartGame();
	if (nextAM != None)
		bAdvancedResult = nextAM.HandleRestartGame();
	return (bNormalResult || bAdvancedResult);
}

function bool HandleEndGame() {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.HandleEndGame();
	if (nextAM != None)
		bAdvancedResult = nextAM.HandleEndGame();
	return (bNormalResult || bAdvancedResult);
}

function bool HandlePickupQuery(Pawn Other, Inventory item, out byte bAllowPickup) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.HandlePickupQuery(Other, item, bAllowPickup);
	
	if (nextAM != None)
		bAdvancedResult = nextAM.HandlePickupQuery(Other, item, bAllowPickup);
	
	return (bNormalResult || bAdvancedResult);
}

function bool PreventDeath(Pawn Killed, Pawn Killer, name damageType, vector HitLocation) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.PreventDeath(Killed, Killer, damageType, HitLocation);
	if (nextAM != None)
		bAdvancedResult = nextAM.PreventDeath(Killed, Killer, damageType, HitLocation);
	return (bNormalResult || bAdvancedResult);
}

function ModifyLogin(out class<playerpawn> SpawnClass, out string Portal, out string Options) {
	super.ModifyLogin(SpawnClass, Portal, Options);
	if (nextAM != None)
		nextAM.ModifyLogin(SpawnClass, Portal, Options);
}

function ScoreKill(Pawn Killer, Pawn Other) {
	super.ScoreKill(Killer, Other);
	if (nextAM != None)
		nextAM.ScoreKill(Killer, Other);
}

//FIXME
function Class<Weapon> MutatedDefaultWeapon() {
	local Class<Weapon> result;
	
	result = super.MutatedDefaultWeapon();
	if (nextAM != None)
		result = nextAM.MutatedDefaultWeapon();
	return result;
}

//FIXME
function Class<Weapon> MyDefaultWeapon() {
	local Class<Weapon> result;
	
	result = super.MyDefaultWeapon();
	if (nextAM != None)
		result = nextAM.MyDefaultWeapon();
	return result;
}

function bool ReplaceWith(actor Other, string aClassName) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.ReplaceWith(Other, aClassName);
	if (nextAM != None)
		bAdvancedResult = nextAM.ReplaceWith(Other, aClassName);
	return (bNormalResult || bAdvancedResult);
}

function bool AlwaysKeep(Actor Other) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.AlwaysKeep(Other);
	if ((nextAM != None) && (!bNormalResult))
		bAdvancedResult = nextAM.AlwaysKeep(Other);
	return (bNormalResult || bAdvancedResult);
}

function bool IsRelevant(Actor Other, out byte bSuperRelevant) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.IsRelevant(Other, bSuperRelevant);
	if (nextAM != None)
		bAdvancedResult = nextAM.IsRelevant(Other, bSuperRelevant);
	return (bNormalResult || bAdvancedResult);
}

function bool CheckReplacement(Actor Other, out byte bSuperRelevant) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.CheckReplacement(Other, bSuperRelevant);
	if (nextAM != None)
		bAdvancedResult = nextAM.CheckReplacement(Other, bSuperRelevant);
	//If at least one mutator returns true, the actor will not be removed, if all the mutators return true, the actor will be removed.
	return (bNormalResult || bAdvancedResult);
}

function Mutate(string MutateString, PlayerPawn Sender) {
	local ActorInfo AI;
	local String param1, param2, param3, param4, param5, param6;
	local bool bScreenLogs;
	local ValueRegister vReg;

	super.Mutate(MutateString, Sender);
	if (nextAM != None)
		if ((MutateString != "ViewNextAI") && (Left(MutateString, 13) != "ViewAIByOwner") && (Left(MutateString, 17) != "ViewNextAIByClass") && (Left(MutateString, 8) != "LogChain") && (Left(MutateString, 3) != "get") && (Left(MutateString, 3) != "set"))
			nextAM.Mutate(MutateString, Sender);
	
	if (bDebug) {
		param1 = FindSplittedStr(MutateString, " ", 1);
		param2 = FindSplittedStr(MutateString, " ", 2);
		param3 = FindSplittedStr(MutateString, " ", 3);
		param4 = FindSplittedStr(MutateString, " ", 4);
		param5 = FindSplittedStr(MutateString, " ", 5);
		param6 = FindSplittedStr(MutateString, " ", 6);
	
		if (MutateString ~= "ViewNextAI") {
			Sender.ConsoleCommand("cheatview ActorInfo");
			if (Sender.ViewTarget.IsA('ActorInfo'))
				Sender.ClientMessage("[AI:"@Sender.ViewTarget.Owner$"] [Owner:"@ActorInfo(Sender.ViewTarget).Owner$"]");
		} else if (Left(MutateString, 14) ~= "ViewAIByOwner ") {
			if (Len(MutateString) > 14) {
				AI = getActorAI(getActorFromText(param1, true));
				if (AI != None) {
					Sender.ViewTarget = AI;
					Sender.bBehindview = true;
					Sender.ClientMessage(AI);
				} else {
					Sender.ClientMessage("Failed to find the AI of "$param1$".");
				}
			}
		} else if (Left(MutateString, 18) ~= "ViewNextAIByClass ") {
			if (Len(MutateString) > 18) {
				Sender.ConsoleCommand("cheatview "$param1);
				Sender.ClientMessage(getActorAI(Sender.ViewTarget));
			}
		} else if (Left(MutateString, 9) ~= "LogChain ") {
			bScreenLogs = bool(param4);
			
			LogChain(getActorFromText(param1), param2, param3, bScreenLogs);
		} else if ((Left(MutateString, 4) ~= "get ") || (Left(MutateString, 4) ~= "set ")) {
			AI = getActorAI(getActorFromText(param1));
			
			if (Left(MutateString, 3) ~= "get") {
				getRegisteredValue(AI, param2, int(param3), vReg);
				if (vReg != None)
					Level.Game.BroadcastMessage("[PROPERTY:"@vReg.propertyName$"] [VALUE:"@vReg.propertyValue$"] [CHANNEL:"@vReg.channel$"] [bCheckStatus:"@vReg.bCheckStatus$"] [bUnchanged:"@vReg.bUnchanged$"] [bUnchangedLoop:"@vReg.bUnchangedLoop$"] [lastChangeTime:"@vReg.lastChangeTime$"] [bRealValue:"@vReg.bRealValue$"]");
				else
					Level.Game.BroadcastMessage("Cannot find or register a value with this name for the specified actor."); //"get" is called from "set" too.
			} else {
				if (AI != None) {
					//Check if the actor already has a variable with this valueName.
					getRegisteredValue(AI, param2, int(param6), vReg);
					if (vReg != None) {
						vReg.propertyValue = param3;
						Level.Game.BroadcastMessage("The value has been updated for the specified actor.");
					} else {
						registerTextValue(AI, param2, param3, bool(param4), bool(param5), int(param6));
					
						//Check if success.
						Level.Game.BroadcastMessage("The new value has been registered for the specified actor.");
						Sender.ConsoleCommand("mutate get"@param1@param2@param6);
					}
				}
			}
		}
	}
}

function MutatorTakeDamage( out int ActualDamage, Pawn Victim, Pawn InstigatedBy, out Vector HitLocation, 
						out Vector Momentum, name DamageType) {
	super.MutatorTakeDamage( ActualDamage, Victim, InstigatedBy, HitLocation, Momentum, DamageType );
		
	if (nextDamageAM != None)
		nextDamageAM.MutatorTakeDamage( ActualDamage, Victim, InstigatedBy, HitLocation, Momentum, DamageType );
}

function bool MutatorTeamMessage( Actor Sender, Pawn Receiver, PlayerReplicationInfo PRI, coerce string S, name Type, optional bool bBeep ) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.MutatorTeamMessage( Sender, Receiver, PRI, S, Type, bBeep );
		
	if (nextMessageAM != None)
		bAdvancedResult = nextMessageAM.MutatorTeamMessage( Sender, Receiver, PRI, S, Type, bBeep );
	return (bNormalResult || bAdvancedResult);
}

function bool MutatorBroadcastMessage( Actor Sender, Pawn Receiver, out coerce string Msg, optional bool bBeep, out optional name Type ) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.MutatorBroadcastMessage( Sender, Receiver, Msg, bBeep, Type );
		
	if (nextMessageAM != None)
		bAdvancedResult = nextMessageAM.MutatorBroadcastMessage( Sender, Receiver, Msg, bBeep, Type );
	return (bNormalResult || bAdvancedResult);
}

function bool MutatorBroadcastLocalizedMessage( Actor Sender, Pawn Receiver, out class<LocalMessage> Message, out optional int Switch, out optional PlayerReplicationInfo RelatedPRI_1, out optional PlayerReplicationInfo RelatedPRI_2, out optional Object OptionalObject ) {
	local bool bNormalResult, bAdvancedResult;
	
	bNormalResult = super.MutatorBroadcastLocalizedMessage( Sender, Receiver, Message, Switch, RelatedPRI_1, RelatedPRI_2, OptionalObject );
		
	if (nextMessageAM != None)
		bAdvancedResult = nextMessageAM.MutatorBroadcastLocalizedMessage( Sender, Receiver, Message, Switch, RelatedPRI_1, RelatedPRI_2, OptionalObject );
	return (bNormalResult || bAdvancedResult);
}

simulated function RegisterHUDMutator() {
	local Pawn P;
	local Mutator HUDMut;
	local bool bAdded;

	for (P = level.PawnList; P != None; P = P.NextPawn)
		if (P.IsA('PlayerPawn') && (PlayerPawn(P).myHUD != None)) {
				for (HUDMut = PlayerPawn(P).myHUD.HUDMutator; HUDMut != None; HUDMut = HUDMut.NextHUDMutator) {
					if (HUDMut.IsA('AdvancedMutator')) {
						nextHUDAM = AdvancedMutator(HUDMut).nextHUDAM;
						AdvancedMutator(HUDMut).nextHUDAM = Self;
						
						bAdded = true;
						break;
					}
				}
				
				if (!bAdded)
					super.RegisterHUDMutator();
				
			bHUDMutator = True;
		}
}

////////////////////////////////////////////////////////////////////////////////
//Testing functions.

function LogChain(Actor actor, String actorChainName, String blockChainName, optional bool bScreenLogs) {
	local Actor firstBlock, nextBlock;
	local int blockNum;
	
	if ((actor != None) && (actorChainName != "") && (blockChainName != "")) {
		firstBlock = getActorFromProperty(actor, actorChainName);
		
		if (firstBlock == None) {
			if (!bScreenLogs)
				Log("LogChain:"@"Chain is empty!");
			else
				Level.Game.BroadcastMessage("LogChain:"@"Chain is empty!");
		} else {
			for (nextBlock = firstBlock; nextBlock != None; nextBlock = getActorFromProperty(nextBlock, blockChainName)) {
				if (!bScreenLogs)
					Log("ChainLog:"@"["$blockNum$"]"@nextBlock);
				else
					Level.Game.BroadcastMessage("ChainLog:"@"["$blockNum$"]"@nextBlock);
					
				blockNum++;
			}
		}
	} else {
		if (!bScreenLogs)
			Log("LogChain:"@"Chain cannot be logged!");
		else
			Level.Game.BroadcastMessage("LogChain:"@"Chain cannot be logged!");
	}
}

defaultproperties {
	maxNoneTimes=3
}