//=============================================================================
// AdvancedMutatorManager.
// 
// Date: 2016
// ©copyright Francesco Biscazzo. All rights reserved.
//=============================================================================
class AdvancedMutatorManager extends Info;

//AdvancedMutator chains.
var AdvancedMutator BaseAM;
var AdvancedMutator DamageAM;
var AdvancedMutator MessageAM;

var bool bFirstIteration;
var Actor lastAI, lastNoDelete; //Used to not to repeat the check to the whole actors in the map.

var bool bRestartNotified;

function RegisterAdvancedMutator(AdvancedMutator AdvMut) {
	BaseAM = AdvMut;
	bFirstIteration = true;
}

function Tick (float Delta) {
	local Actor a;
	local bool lastReached;
	local ActorInfo AI;
	local Mutator mut;

	super.Tick(Delta);
	
	//Give the AI only to the newly spawned actors.
	
	//The AllActors iterator iterates throught a chain of Actors where actors are added to the back of the chain.
	if (BaseAM != None) {
		foreach AllActors(class'Actor', a) {
			if (bFirstIteration || lastReached) {
				if (a.Instigator != None)
					//If the actor is a spawned one, let mutators handle it.
					Level.Game.IsRelevant(a);
					
				if (BaseAM.startWithAI(a)) {
					AI = GiveAI(a);
					if (AI != None) {
						AI.Activate();
						lastAI = AI;
					}
				}
				
				if (a.bNoDelete)
					lastNoDelete = a;
			} else {
				//Let the check start from the last loaded non-None actor.
				if (lastAI != None) {
					//Best case.
					if (a == lastAI)
						lastReached = true;
				} else if (lastNoDelete != None) {
					//Worste case.
					if (a == lastNoDelete)
						lastReached = true;
				} else {
					bFirstIteration = true;
				}
			}
		}
		
		bFirstIteration = false;
	}
	
	//If a new mutator is registered as damage/message mutator, if it's an AdvancedMutator put it in the forked chain.
	mut = Level.Game.DamageMutator;
	if ((mut != None) && mut.IsA('AdvancedMutator')) {
		RegisterDamageAM(AdvancedMutator(mut));
	}
	mut = Level.Game.MessageMutator;
	if ((mut != None) && mut.IsA('AdvancedMutator')) {
		RegisterMessageAM(AdvancedMutator(mut));
	}
	
	//Call HandleRestartGame() to all the active mutators if the level is restarting.
	if (!bRestartNotified && (Level.NextURL ~= "?Restart")) {
		Level.Game.BaseMutator.HandleRestartGame();
		bRestartNotified = true;
	}
}

function RegisterDamageAM(AdvancedMutator AM)
{
	AM.NextDamageAM = DamageAM;
	DamageAM = AM;
}

function RegisterMessageAM(AdvancedMutator AM)
{
	AM.NextMessageAM = MessageAM;
	MessageAM = AM;
}

function ActorInfo GiveAI(Actor actor) {
	local ActorInfo newAI;
	
	if (actor != None) {
		if (BaseAM.canHaveAI(actor)) {
			newAI = actor.Spawn(class'ActorInfo', actor);
			if (newAI != None)
				newAI.RegisterAdvancedMutator(Self);
		}
	}
	
	return newAI;
}

//Give an AI to any actor of the specified class.
function bool GiveAIToClass(class<Actor> actorClass, optional bool bNoSubclass) {
	local Actor a;
	local bool bResult;

	if (actorClass != None)
		foreach AllActors(actorClass, a) {
			if (bool(BaseAM.static.ClassIsRelationedWith(a.class, actorClass, bNoSubclass)))
				if (GiveAI(a) != None)
					bResult = true;
		}
	
	return bResult;
}

//Remove a possible AI from any actor of the specified class.
function RemoveAIFromClass(class<Actor> actorClass, optional bool bNoSubclass) {
	local Actor a;
	local bool bResult;
	local ActorInfo AI;

	if (actorClass != None)
		foreach AllActors(actorClass, a) {
			if (bool(BaseAM.static.ClassIsRelationedWith(a.class, actorClass, bNoSubclass))) {
				AI = BaseAM.getActorAI(a);
				if (AI != None)
					AI.Destroy();
			}
		}
}

//Activate/Deactivate a possible AI for any actor of the specified class.
function bool ToggleAIOfClass(class<Actor> actorClass, optional bool bDeactivate, optional bool bNoSubclass) {
	local ActorInfo AI;
	local bool bResult;

	if (actorClass != None)
		foreach AllActors(class'ActorInfo', AI) {
			if (AI.Owner != None) {
				if (bool(BaseAM.static.ClassIsRelationedWith(AI.Owner.class, actorClass, bNoSubclass))) {
					if (bDeactivate)
						AI.Deactivate();
					else
						AI.Activate();
				}
				
				bResult = true;
			}
		}
	
	return bResult;
}

//Called when a registered value changes.
function BroadcastValueStatusChanged(class<AdvancedMutator> requesterClass, Actor actor, String propertyName, String oldValue, String newValue, int channel) {
	local AdvancedMutator nextAM;
	
	for (nextAM = BaseAM; nextAM != None; nextAM = nextAM.nextAM)
		if (bool(BaseAM.static.ClassIsRelationedWith(requesterClass, nextAM.class)))
			nextAM.ValueStatusChanged(actor, propertyName, oldValue, newValue, channel);
}

//Called when a registered value has been unchanged for one tick.
//NOTE: Called once only after the first unchanging if bUnchangedLoop is set to false.
function BroadcastValueStatusUnchanged(class<AdvancedMutator> requesterClass, Actor actor, String propertyName, String value, float lastChangeTime, int channel) {
	local AdvancedMutator nextAM;
	
	for (nextAM = BaseAM; nextAM != None; nextAM = nextAM.nextAM)
		if (bool(BaseAM.static.ClassIsRelationedWith(requesterClass, nextAM.class)))
			nextAM.ValueStatusUnchanged(actor, propertyName, value, lastChangeTime, channel);
}